//--------------------------------------------------------------------------------
//
//                    CX5530 - Audio functions
//
//     This is a group of hardware specific functions that define a general 
//  interface to be portable to other Operating systems and to run on the 
//  National Semiconductor/Cyrix MediaGX CPU and 5530. 
//
//     These functions include interfaces to an AC'97 Codec, Hardware 
//  Initialization and DMA buffer input/Output and currently run on 
//  Unicorn and Scorpius platforms using LM4548 Codecs.
//
//  OS_inc.h - Contains the OS-Specific definitions/Macros
//             The driver developer has to define these macros according to the O/S
//             to which the driver is being developed.
//--------------------------------------------------------------------------------

#include "duraudio.h" 
#include "os_inc.h" 

const unsigned short VolumeLUT[] = {
    0x1f, 0x1e, 0x1d, 0x1c, 0x1b, 0x1a, 0x19, 0x18,
    0x17, 0x16, 0x15, 0x14, 0x13, 0x12, 0x11, 0x10, 
    0x0f, 0x0e, 0x0d, 0x0c, 0x0b, 0x0a, 0x09, 0x08,
    0x07, 0x06, 0x05, 0x04, 0x03, 0x02, 0x01, 0x00 };

const struct
{
	char			*Name;
	unsigned long	ID;
	unsigned char	fCS5535;
} Device[] =  { 
	{ "Centaurus",	 0x01001078, FALSE }, 
	{ "Scorpius",	 0x04001078, FALSE }, 
	{ "Carmel",		 0x0500100b, FALSE }, 
	{ "Tavor",		 0x0510100b, FALSE },
	{ "Hawk",		 0x002B100b, TRUE  }, 
	{ "",			 0x00000000, FALSE }}; 


unsigned long DMA_buffers_phys_addr;

//
// the following arrays contains 2 entries, One for input, one for output
//
durLPWAVEFORMATEX g_pwfx[2];

/*
static void (*pfnFillBuffer)(PDURAUDIO pGeode, durPWAVEHDR, unsigned char *);
static void (*pfnGetBuffer) (PDURAUDIO pGeode, durPWAVEHDR, unsigned char *);

static unsigned char *dma_out_page[AUDIO_DMA_PAGES];
static unsigned char *dma_in_page [AUDIO_DMA_PAGES];
*/

//-----------------------------------------------------------------------------
//
//  PDURAUDIO DURAUDIO_Initialize (unsigned int Irq)
//
//  Initializes the 5530 hardware.
//  It has to be called with the IRQ assigned by the OS
//
//  This function:
//
//  - Reset and initialize the CODEC 
//	
//--------------------------------------------------------------------------------
PDURAUDIO DURAUDIO_Initialize (unsigned int Irq)
{
	PDURAUDIO	pGeode;
	
	OS_DbgMsg("DURAUDIO_Initialize\n");

	pGeode=(PDURAUDIO) OS_AllocateMemory(sizeof(DURAUDIO));

	if (pGeode==NULL)
		return FALSE;

	memset (pGeode, 0, sizeof(DURAUDIO));

	if(!DURAUDIO_InitAudioRegs(pGeode))
		return FALSE;

	if (Irq!=POLLING)
	{
		pGeode->fPolling=FALSE;

		DURAUDIO_SetupIRQ (pGeode, (unsigned long) Irq);

		//
		//   Save the selected IRQ
		//
		pGeode->SelectedIRQ=Irq;

		// 
		//  Prepare the bit mask (To clear Interrupts)
		//
		pGeode->uIRQMask=1;
		pGeode->uIRQMask <<= pGeode->SelectedIRQ;
		pGeode->uIRQMask=~pGeode->uIRQMask;
	}
	else
	{
		pGeode->fPolling=TRUE;
	}

	//
    // Setup Flags for input and output buffers
    //
    //   v_nNextPage:  Points to the next buffer to be filled with data.
    //   v_fMoreData:  Flag that signals if there is more data coming.
    //
    pGeode->AudioChannel[WAPI_OUT].v_nNextPage = 0;
    pGeode->AudioChannel[WAPI_OUT].v_fMoreData = FALSE;

    pGeode->AudioChannel[WAPI_IN].v_nNextPage  = 0;
    pGeode->AudioChannel[WAPI_IN].v_fMoreData  = FALSE;

	return pGeode;
}

//-----------------------------------------------------------------------------
//
//  void DURAUDIO_SetupIRQ (unsigned long Irq)
//
//  - Programs the IRQ 
//  - Sets VSA audio irq component to create the Virtual interrupt (CS5530)
//
//-----------------------------------------------------------------------------
void DURAUDIO_SetupIRQ (PDURAUDIO pGeode, unsigned long Irq)
{
	if (!pGeode->fCS5535)
	{
		unsigned long	SetIRQData;

		//
		//   If it is a CS5530, we have to configure VSA...
		//

		//
		// VSA2 IRQ config method
		//
		OS_WritePortUShort( 0xAC1C, 0xFC53 );					// Unlock virtual registers
		OS_WritePortUShort( 0xAC1C, 0x108 );						// ID the audio config virtual regs
		OS_WritePortUShort( 0xAC1E, (unsigned short) Irq );		// Set the value

		//
		// VSA1 IRQ config method
		//

		//
		//  Get the Irq and OR it with the command
		//
		SetIRQData = ((Irq<<16) | 0xA00A);

		OS_DbgMsg("SetIRQData: %08X ", SetIRQData);

		OS_WritePortULong((unsigned short) PCI_CADDR, 0x800090D0);		//  Set the address
		OS_WritePortULong((unsigned short) PCI_CDATA, SetIRQData);		//  Send the command

		//
		//   Set the InterruptID address for the CS5530
		//
		pGeode->pInterruptID = (unsigned char *) OS_VirtualAddress(0x000004F0);
	}
}

unsigned long DURAUDIO_InitAudioRegs(PDURAUDIO pGeode)
{
	unsigned long   F3BAR0_Phys, config_data, pci_config_addr, Temp;       
	unsigned short  i, j, pci_cmd;      
	
    //
    //        PCI Audio/DMA  init
    //
    //   Go probe for PCI device reg spaces...
    //   NOTE: This loop checks that the 5530 Bridge Header is found.
	//

	for(i=0;i<32;i++) 
	{
		//
        // Iterate over all dev numbers in PCI Config addr reg.
		//
		pci_config_addr = 0x80000000 | (i<<11);  

		//
        // Get Vendor_ID+dev_id reg (1st word in PCI Header). 
		//
		OS_WritePortULong( PCI_CADDR, pci_config_addr );  
		config_data = OS_ReadPortULong( PCI_CDATA ); 
		
		//
        // Check if the device is one of ours...
		//
		for (j=0; Device[j].ID; ++j)
		{
			if (Device[j].ID==config_data)
				goto Found_Device;
		}
	}
	
    //
    //  Didn't find the Device ID.  Return.
    //
    OS_DbgMsg("Cant find PCI Audio dev!\n");
	return ( FALSE );
	
    //    
    //   Found the device ID.  Now initialize the device.
    //
	
    Found_Device:;
	
	OS_DbgMsg("Found Device: %s\n", Device[j].Name);

	pGeode->fCS5535=Device[j].fCS5535;

	//
    // 10-13h. BASE REGISTER --> VSA Audio Regs are mem mapped at this loc.
	//
	pGeode->PCI_Header_Registers = pci_config_addr | PCI_FUNC3_AUDIO;
	F3BAR0_Phys = DURAUDIO_IO_Read32( pGeode->PCI_Header_Registers, 0x10 );

    //
    // Read the PCI Command register
    //
	pci_cmd = DURAUDIO_IO_Read16( pGeode->PCI_Header_Registers, PCI_AUDIO_CMD_REG );

	if (F3BAR0_Phys & 0x00000001) 
	{
		pGeode->fIOAccess=TRUE;

		//
		//   Set it to IO mapping
		//
		pci_cmd |= 0x05;   
		DURAUDIO_IO_Write16( pGeode->PCI_Header_Registers, PCI_AUDIO_CMD_REG, pci_cmd );   

		pGeode->F3BAR0 = F3BAR0_Phys & 0xFFFFFFFE;
	}
	else
	{
		pGeode->fIOAccess=FALSE;

		//
		//   Set it to memory mapping
		//
		pci_cmd |= 0x06;   
		DURAUDIO_IO_Write16( pGeode->PCI_Header_Registers, PCI_AUDIO_CMD_REG, pci_cmd );   

		pGeode->F3BAR0=OS_Get_F3BAR0_Virt(F3BAR0_Phys);

	}

	if (!pGeode->F3BAR0)
	{
		OS_DbgMsg("Cant Map F3BAR0.\n");
		return ( FALSE );
	}


	//
	//    Initialize the Audio BusMaster register pointers
	//

	if (!pGeode->fPolling)
	{
		pGeode->IRQControlRegister		  = pGeode->F3BAR0+0x1C;
		pGeode->InternalIRQEnableRegister = pGeode->F3BAR0+0x1A;
	}

	pGeode->AudioChannel[CHANNEL0_PLAYBACK].AudioBusMaster.CommandRegister		= pGeode->F3BAR0+0x20;
	pGeode->AudioChannel[CHANNEL0_PLAYBACK].AudioBusMaster.SMI_StatusRegister	= pGeode->F3BAR0+0x21;
	pGeode->AudioChannel[CHANNEL0_PLAYBACK].AudioBusMaster.PRDTableAddress		= pGeode->F3BAR0+0x24;
	pGeode->AudioChannel[CHANNEL0_PLAYBACK].AudioBusMaster.CurrentBufferPointer = pGeode->F3BAR0+0x60;
	pGeode->AudioChannel[CHANNEL0_PLAYBACK].AudioBusMaster.DirectionBit			= PCI_READS;
	pGeode->AudioChannel[CHANNEL0_PLAYBACK].Running=FALSE;

	pGeode->AudioChannel[CHANNEL1_RECORD].AudioBusMaster.CommandRegister		= pGeode->F3BAR0+0x28;
	pGeode->AudioChannel[CHANNEL1_RECORD].AudioBusMaster.SMI_StatusRegister		= pGeode->F3BAR0+0x29;
	pGeode->AudioChannel[CHANNEL1_RECORD].AudioBusMaster.PRDTableAddress		= pGeode->F3BAR0+0x2C;
	pGeode->AudioChannel[CHANNEL1_RECORD].AudioBusMaster.CurrentBufferPointer	= pGeode->F3BAR0+0x64;
	pGeode->AudioChannel[CHANNEL1_RECORD].AudioBusMaster.DirectionBit			= PCI_WRITES;
	pGeode->AudioChannel[CHANNEL1_RECORD].Running=FALSE;

    //
    // CODEC - RESET and volumes initalization.
    //

	if (pGeode->fIOAccess)
	{
		if (pGeode->fCS5535)
		{
			Temp = OS_ReadPortULong ( (unsigned short) (pGeode->F3BAR0 + CODEC_CONTROL_REG_5535) );
			OS_WritePortULong ( (unsigned short) (pGeode->F3BAR0 + CODEC_CONTROL_REG_5535), (Temp | 0x00020000) );

			if (!DURAUDIO_WaitForBit (pGeode, CODEC_STATUS_REG_5535, BIT_CODEC_READY, SET, 200, NULL))
			{
				OS_DbgMsg("Primary Codec NOT Ready...Aborting\n");
				return FALSE;
			}
		}
		else
		{
			OS_WritePortULong ( (unsigned short) (pGeode->F3BAR0 + CODEC_CMD_REG), 0L );
		}
	}
	else
	{
		*((unsigned long *)( pGeode->F3BAR0 + CODEC_CMD_REG )) = 0L;
	}

	//
	//  Check which codec is being used 
	//
	if (DURAUDIO_CodecRead(pGeode, AD1819A_VENDORID1) == 0x4144 &&
	    DURAUDIO_CodecRead(pGeode, AD1819A_VENDORID2) == 0x5303) 
	{
		pGeode->fAD1819A = TRUE;
		//  Enable non-48kHz sample rates. 
		DURAUDIO_CodecWrite (pGeode, AD1819A_SER_CONF, (unsigned short) (DURAUDIO_CodecRead(pGeode, AD1819A_SER_CONF>>8) | AD1819A_SER_CONF_DRQEN));
	}
	else
	{
		pGeode->fAD1819A = FALSE;
		DURAUDIO_CodecWrite (pGeode, EXT_AUDIO_CTRL_STAT, (unsigned short) (DURAUDIO_CodecRead(pGeode,EXT_AUDIO_CTRL_STAT) | 0x0001));	// set the VRA bit to ON
	}

    // set default volume
    DURAUDIO_CodecWrite(pGeode, MASTER_VOLUME,       0x0a0a);
    DURAUDIO_CodecWrite(pGeode, PCM_OUT_VOL,         0x0000);
    DURAUDIO_CodecWrite(pGeode, PC_BEEP_VOLUME,      0x0000);
    DURAUDIO_CodecWrite(pGeode, PHONE_VOLUME,        0x8000);
    DURAUDIO_CodecWrite(pGeode, MIC_VOLUME,          0x8048);
    DURAUDIO_CodecWrite(pGeode, LINE_IN_VOLUME,      0x0808);
    DURAUDIO_CodecWrite(pGeode, CD_VOLUME,           0x8000);
    DURAUDIO_CodecWrite(pGeode, VIDEO_VOLUME,        0x8000);
    DURAUDIO_CodecWrite(pGeode, TV_VOLUME,           0x8000);
    DURAUDIO_CodecWrite(pGeode, RECORD_SELECT,       0x0000);
    DURAUDIO_CodecWrite(pGeode, RECORD_GAIN,         0x0a0a);
    DURAUDIO_CodecWrite(pGeode, GENERAL_PURPOSE,     0x0200);
    DURAUDIO_CodecWrite(pGeode, MASTER_VOLUME_MONO,  0x0000);

	//
	//   Set all the power state bits to 0 (Reg 26h)
	//
	DURAUDIO_CodecWrite (pGeode, POWERDOWN_CTRL_STAT, 0x0000);

	pGeode->CurrentPowerState=DURAUDIO_D0;

	return TRUE;
}

void DURAUDIO_SetPRDAddress (PDURAUDIO pGeode, unsigned long Channel, unsigned long PRDPhys, unsigned long PRDVirt)
{    
	pGeode->AudioChannel[Channel].PRDPhys=PRDPhys;
	pGeode->AudioChannel[Channel].PRDTable=(PPRD_ENTRY) PRDVirt;
}

void DURAUDIO_SetDMARegs (PDURAUDIO pGeode, unsigned long Channel)
{    
    unsigned char   status;

	if (pGeode->fIOAccess)
	{
		//
		//  Clear the EOP bits
		//
		status = OS_ReadPortUChar ((unsigned short) pGeode->AudioChannel[CHANNEL0_PLAYBACK].AudioBusMaster.SMI_StatusRegister); 
		
		//
		//  Points to PRD array's physical address.
		//
		OS_WritePortULong ((unsigned short) pGeode->AudioChannel[CHANNEL0_PLAYBACK].AudioBusMaster.PRDTableAddress, pGeode->AudioChannel[Channel].PRDPhys);
	}
	else
	{
		//
		//  Clear the EOP bits
		//
		status = *((unsigned char *) pGeode->AudioChannel[CHANNEL0_PLAYBACK].AudioBusMaster.SMI_StatusRegister); 
		
		//
		//  Points to PRD array's physical address.
		//
		*((unsigned long *) pGeode->AudioChannel[CHANNEL0_PLAYBACK].AudioBusMaster.PRDTableAddress) = pGeode->AudioChannel[Channel].PRDPhys; 
	}
}

//-------------------------------------------------------
//  void DURAUDIO_StartDMA (PDURAUDIO pGeode, unsigned long Channel)
//
//  Starts the DMA for the channel specified by Channel.
//
//-------------------------------------------------------
void DURAUDIO_StartDMA (PDURAUDIO pGeode, unsigned long Channel)
{
	if (pGeode->fIOAccess)
	{
		//
		//   Sets the PRD physical address
		//
		OS_WritePortULong ( (unsigned short) pGeode->AudioChannel[Channel].AudioBusMaster.PRDTableAddress, 
							pGeode->AudioChannel[Channel].PRDPhys);
	}
	else
	{
		//
		//   Sets the PRD physical address
		//
		*((unsigned long *) pGeode->AudioChannel[Channel].AudioBusMaster.PRDTableAddress) = pGeode->AudioChannel[Channel].PRDPhys;
	}

	DURAUDIO_ResumeDMA (pGeode, Channel);
	pGeode->AudioChannel[Channel].Running=TRUE;
}

//----------------------------------------------------
//
//  void DURAUDIO_StopDMA (PDURAUDIO pGeode, unsigned long Channel)
//
//  Stops the DMA channel controlled by the specified PRD table
//
//----------------------------------------------------
void DURAUDIO_StopDMA (PDURAUDIO pGeode, unsigned long Channel)
{	
//	unsigned int i;

	if (pGeode->fCS5535)
	{
		if (pGeode->fIOAccess)
		{
			//
			//   Turn OFF DMA
			//
			OS_WritePortUChar ( (unsigned short) pGeode->AudioChannel[Channel].AudioBusMaster.CommandRegister, STOP_BUSMASTER);
		}
		else
		{
			//
			//   Turn OFF DMA
			//
			*((unsigned char *) pGeode->AudioChannel[Channel].AudioBusMaster.CommandRegister) = STOP_BUSMASTER;
		}
	}
	else
	{
		*((unsigned char *) pGeode->AudioChannel[Channel].AudioBusMaster.CommandRegister) = 
					*((unsigned char *) pGeode->AudioChannel[Channel].AudioBusMaster.CommandRegister) & (~ENABLE_BUSMASTER);

/*
		for (i=0; i<(pGeode->AudioChannel[Channel].PRDEntriesAllocated-1); ++i)
		{
			pGeode->AudioChannel[Channel].PRDTable[i].SizeFlags=PRD_EOT_BIT;
		}
		*/
	}
	pGeode->AudioChannel[Channel].Running=FALSE;
}

//----------------------------------------------------
//
//  void DURAUDIO_PauseDMA (PDURAUDIO pGeode, unsigned long Channel)
//
//  Pauses the DMA channel controlled by the specified PRD table
//
//----------------------------------------------------
void DURAUDIO_PauseDMA (PDURAUDIO pGeode, unsigned long Channel)
{	
	if (pGeode->fIOAccess)
	{
		OS_WritePortUChar ( (unsigned short) pGeode->AudioChannel[Channel].AudioBusMaster.CommandRegister, 
							(unsigned char) (pGeode->AudioChannel[Channel].AudioBusMaster.DirectionBit | PAUSE_BUSMASTER));
	}
	else
	{
		*((unsigned char *) pGeode->AudioChannel[Channel].AudioBusMaster.CommandRegister) = pGeode->AudioChannel[Channel].AudioBusMaster.DirectionBit | PAUSE_BUSMASTER;
	}
}

//----------------------------------------------------
//
//  void DURAUDIO_ResumeDMA (PDURAUDIO pGeode, unsigned long Channel)
//
//  Resumes the DMA channel controlled by the specified PRD table
//
//----------------------------------------------------
void DURAUDIO_ResumeDMA (PDURAUDIO pGeode, unsigned long Channel)
{	
	if (pGeode->fIOAccess)
	{
		OS_WritePortUChar ( (unsigned short) pGeode->AudioChannel[Channel].AudioBusMaster.CommandRegister, 
							(unsigned char) (((pGeode->AudioChannel[Channel].AudioBusMaster.DirectionBit) & 0xFC) | ENABLE_BUSMASTER));
	}
	else
	{
		*((unsigned char *) pGeode->AudioChannel[Channel].AudioBusMaster.CommandRegister) = pGeode->AudioChannel[Channel].AudioBusMaster.DirectionBit | ENABLE_BUSMASTER;
	}
}

//---------------------------------------------------------
//
//  void DURAUDIO_CodecWrite( PDURAUDIO pGeode, unsigned short CodecRegister, unsigned short CodecData  )
//
//  Writes data to the CODEC
//
//---------------------------------------------------------
void DURAUDIO_CodecWrite( PDURAUDIO pGeode, unsigned char CodecRegister, unsigned short CodecData  )
{
	int			  i;
	unsigned long CodecRegister_data = 0;
	unsigned long val=0, Temp=0, timeout;
	
	CodecRegister_data = ((unsigned long) CodecRegister)<<24; 
	CodecRegister_data |= (unsigned long) CodecData;
	CodecRegister_data &= CODEC_COMMAND_MASK;
	
	if (pGeode->fCS5535)
	{
		// 
		//  Set the bit.  We are going to access the CODEC...
		// 
		CodecRegister_data |= BIT_5535_CODEC_COMMAND_NEW;

		//
		//  Write the data
		//
		OS_WritePortULong((unsigned short) (pGeode->F3BAR0 + CODEC_CONTROL_REG_5535), CodecRegister_data);
		OS_DbgMsg("Writing: %08X\n", CodecRegister_data);
		
		//
		//  We need to wait for bit16 of the Codec control register to clear
		//
		Temp = OS_ReadPortULong((unsigned short) (pGeode->F3BAR0 + CODEC_CONTROL_REG_5535));
		OS_DbgMsg("Temp: %08X\n", Temp);
	
		timeout = 50;

		while ((Temp & BIT_5535_CODEC_COMMAND_NEW) && timeout-- )
		{
			OS_Sleep(10);
			Temp = OS_ReadPortULong((unsigned short) (pGeode->F3BAR0 + CODEC_CONTROL_REG_5535));
		}

		if (!timeout)
		{
			OS_DbgMsg("Could not Write the CODEC!! BIT_5535_CODEC_COMMAND_NEW did not clear!\n");
		}
	}
	else
	{
		//
		//	Wait for the right time to write
		//
		DURAUDIO_WaitFrameAndWrite( pGeode, CodecRegister_data );
		
		//
		//   Wait for Status Tag and Status Valid bits in the CODEC Status Register
		//
		for( i=0; i<=30000; i++) 
		{
			if (pGeode->fIOAccess)
			{
				val = OS_ReadPortULong ( (unsigned short) (pGeode->F3BAR0 + CODEC_STATUS_REG) );
			}
			else
			{
				val = *((unsigned long *)( pGeode->F3BAR0 + CODEC_STATUS_REG ));
			}
			
			if( ( val & CODEC_STATUS_VALID ) || ( val & CODEC_STATUS_NEW )) 
				break;
		} 
		
		// This prevents multiple volume writes to a codec register 
		DURAUDIO_WaitFrameAndWrite( pGeode, 0x80000000 | CodecRegister_data );
	}
}

unsigned short DURAUDIO_CodecRead ( PDURAUDIO pGeode, unsigned char CodecRegister )
{
	int			  i;
	unsigned long CodecRegister_data = 0;
	unsigned long timeout=10;
	volatile unsigned long val=0;
	
	CodecRegister_data  = ((unsigned long)CodecRegister)<<24;
	CodecRegister_data |= 0x80000000;   // High-bit set (p.106) is a CODEC reg READ.

	if (pGeode->fCS5535)
	{
		//
		//  Set the bit.  We are going to access the CODEC...
		//
		CodecRegister_data |= BIT_5535_CODEC_COMMAND_NEW;

		//
		//  Request the data
		//
		OS_WritePortULong((unsigned short) (pGeode->F3BAR0 + CODEC_CONTROL_REG_5535), CodecRegister_data);

		//
		//  Now we need to wait for BIT_5535_CODEC_COMMAND_NEW of the Codec control register to clear
		//  (For subsequent Reads/Writes)
		//
		if (!DURAUDIO_WaitForBit (pGeode, CODEC_CONTROL_REG_5535, BIT_5535_CODEC_COMMAND_NEW, CLEAR, 500, NULL))
		{
			OS_DbgMsg("BIT_5535_CODEC_COMMAND_NEW did not clear!!\n");
		}

		//
		// Wait for CODEC_STATUS_NEW and confirm the read of the requested register
		//
		timeout = 50;
		do
		{
			if (pGeode->fIOAccess)
				val = OS_ReadPortULong((unsigned short) (pGeode->F3BAR0 + CODEC_STATUS_REG_5535));
			else
				val = *((unsigned long *)( pGeode->F3BAR0 + CODEC_STATUS_REG_5535 ));

			if ((val & BIT_5535_CODEC_STATUS_NEW) && ((unsigned long) CodecRegister == ((0xFF000000 & val)>>24)))
			{
				break;
			}
			else
			{
				//
				//   Wait for 10 miliseconds and try again
				//
				OS_Sleep(10);
			}
		} while ( --timeout);

		if (!timeout)
		{
			OS_DbgMsg("Could not read the CODEC!!  Returning what we got.\n");
		}

	}
	else
	{
		do 
		{
			DURAUDIO_WaitFrameAndWrite (pGeode, CodecRegister_data);

			//
			//   Wait for Status Tag and Status Valid bits in the CODEC Status Register
			//
			for( i=0; i<=3000; i++) 
			{
				if (pGeode->fIOAccess)
				{
					val = OS_ReadPortULong((unsigned short) (pGeode->F3BAR0 + CODEC_STATUS_REG));
				}
				else
				{
					val = *((unsigned long *)( pGeode->F3BAR0 + CODEC_STATUS_REG ));
				}
			} 
		} while ((((unsigned long) (0xFF & CodecRegister)) != ((0xFF000000 & val)>>24)) && (--timeout) ); // Check if the register read is the one we want
	}

	return( (unsigned short)val );
}

unsigned char DURAUDIO_WaitForBit(PDURAUDIO		pGeode, 
								  unsigned long Offset, 
								  unsigned long Bit, 
								  unsigned char Operation,
								  unsigned long timeout, 
								  unsigned long *pReturnValue)
{
	volatile unsigned long	Temp;

	if (pGeode->fIOAccess)
		Temp = OS_ReadPortULong((unsigned short) (pGeode->F3BAR0 + Offset));
	else
		Temp = *((unsigned long *)( pGeode->F3BAR0 + Offset ));

	while (timeout)
	{
		if (Operation==CLEAR)
		{
			if (!(Temp & Bit))
				break;
		}
		else
		{
			if (Temp & Bit)
				break;
		}
		//
		//  If the Bit is not clear yet, we wait for 1 milisecond and try again
		//
		OS_Sleep(1);

		if (pGeode->fIOAccess)
			Temp = OS_ReadPortULong((unsigned short) (pGeode->F3BAR0 + Offset));
		else
			Temp = *((unsigned long *)( pGeode->F3BAR0 + Offset ));
 
		timeout--;
	};

	if (pReturnValue)
		*pReturnValue=Temp;

	if (!timeout)
	{
		return FALSE;
	}
	
	return TRUE;
}

//------------------------------------------------------------------------------
//
//  void DURAUDIO_WaitFrameAndWrite(unsigned long val)
//
//  Waits for next avail pos in the CODEC's serial frame and sends the data.
//
//  While the CODEC_CMD_VALID bit is set, it should not be overwritten.
//  i.e., the codec cycles around to ship the command out at the next 
//  avail pos in the serial frame.
//
//------------------------------------------------------------------------------

void DURAUDIO_WaitFrameAndWrite (PDURAUDIO pGeode, unsigned long val)
{
	unsigned long    cmd_val, timeout;
	
	timeout = 30000;
	
	if (pGeode->fIOAccess)
	{
		if (pGeode->fCS5535)
		{
			while ((OS_ReadPortULong((unsigned short) (pGeode->F3BAR0+CODEC_CMD_REG)) & BIT_5535_CODEC_COMMAND_NEW) && (--timeout));
		}
		else
		{
			while ((OS_ReadPortULong((unsigned short) (pGeode->F3BAR0+CODEC_CMD_REG)) & CODEC_CMD_VALID) && (--timeout));
		}

		cmd_val = val & CODEC_COMMAND_MASK;	
		OS_WritePortULong ((unsigned short) (pGeode->F3BAR0+CODEC_CMD_REG), cmd_val);
	}
	else
	{
		while ((*((unsigned long *)( pGeode->F3BAR0 + CODEC_CMD_REG )) & CODEC_CMD_VALID) && (--timeout));

		cmd_val = val & CODEC_COMMAND_MASK;
		*((unsigned long *)( pGeode->F3BAR0 + CODEC_CMD_REG )) = cmd_val;
	}
}

//--------------------------------------------------------------
//   void DURAUDIO_SetCodecRate(PDURAUDIO pGeode, unsigned long Channel, unsigned long Rate)
//  
//   Sets the CODEC Sample Rate
//---------------------------------------------------------------
void DURAUDIO_SetCodecRate(PDURAUDIO pGeode, unsigned long Channel, unsigned long SampleRate)
{
	unsigned short val;

	OS_DbgMsg ("Rate: %d\n", SampleRate);

	pGeode->AudioChannel[Channel].SampleRate=SampleRate;

	//
	//  If Double-Rate is supported (Bit 2 on register 28h)...
	//
	val=DURAUDIO_CodecRead(pGeode, EXTENDED_AUDIO_ID);

	if (val & 0x02)
	{
		OS_DbgMsg ("Codec supports Double rate.\n");
		val=DURAUDIO_CodecRead(pGeode, EXT_AUDIO_CTRL_STAT);

		if (SampleRate>48000)
		{
			DURAUDIO_CodecWrite(pGeode, EXT_AUDIO_CTRL_STAT, (unsigned short) (val|0x0002));
			SampleRate/=2;
		}
		else
			DURAUDIO_CodecWrite(pGeode, EXT_AUDIO_CTRL_STAT, (unsigned short) (val&0xFFFD));
	}

	if (pGeode->fAD1819A) 
	{
		OS_DbgMsg ("AD1819...\n");
		if (Channel)
			DURAUDIO_CodecWrite(pGeode, AD1819A_PCM_SR1,(unsigned short) SampleRate);
		else
			DURAUDIO_CodecWrite(pGeode, AD1819A_PCM_SR0,(unsigned short) SampleRate);
	} 
	else 
	{
		if (Channel)
			DURAUDIO_CodecWrite(pGeode, PCM_LR_ADC_RATE,   (unsigned short) SampleRate);
		else
			DURAUDIO_CodecWrite(pGeode, PCM_FRONT_DAC_RATE,(unsigned short) SampleRate);
	}

	val=DURAUDIO_CodecRead(pGeode, PCM_FRONT_DAC_RATE);
	OS_DbgMsg ("Sample rate in the CODEC (%s): %04X - %dHz\n", Channel?"CAPTURE":"PLAYBACK", val, val);
}

//----------------------------------------------------------
// unsigned long DURAUDIO_SetPowerState(PDURAUDIO pGeode, DURAUDIO_POWER_STATE NewPowerState)
//
//  Sets the power state of the Codec
//
//----------------------------------------------------------
unsigned long DURAUDIO_SetPowerState(PDURAUDIO pGeode, DURAUDIO_POWER_STATE NewPowerState)
{
	if (NewPowerState==pGeode->CurrentPowerState)
		return DURAUDIO_SUCCESS;

	switch (NewPowerState)
	{
		case DURAUDIO_D0:
		{
			//
			//    If coming from D4 (D2, D3 or D4), RESET codec and wait 50ms
			//
            if (pGeode->CurrentPowerState == DURAUDIO_D4)
			{
				DURAUDIO_DisableSMIs (pGeode);
				DURAUDIO_ResetCodec  (pGeode);

				OS_Sleep(50);

				//
				//  Set the codec to Full Power
				//
				if (!DURAUDIO_CodecFullOn (pGeode))
				{
					return	DURAUDIO_ERROR;
				}

				//
				// If we are coming from D2, D3 or D4 we have to restore the registers cause
				// there might have been a power loss.
				//

				if (pGeode->CurrentPowerState == DURAUDIO_D4)
				{
					DURAUDIO_RestoreAudioContext (pGeode);
				}
			}
			else
			{
				//
				//  If we are coming from D1, we just set the codec to Full Power
				//
				if (!DURAUDIO_CodecFullOn (pGeode))
				{
					return	DURAUDIO_ERROR;
				}
			}     

			pGeode->CurrentPowerState=DURAUDIO_D0;

			break;
		}

		case DURAUDIO_D1:
		{
            if (pGeode->CurrentPowerState == DURAUDIO_D4)
            {
				//
				//    If coming from D4 (D2, D3 or D4), RESET codec, wait 50ms, 
				//    Set Codec to Full power state and Restore the registers
				//
				DURAUDIO_DisableSMIs (pGeode);
				DURAUDIO_ResetCodec  (pGeode);

				OS_Sleep(50);

				if (!DURAUDIO_CodecFullOn (pGeode))
				{
					return	DURAUDIO_ERROR;
				}

				DURAUDIO_RestoreAudioContext (pGeode);
            }

			if (pGeode->AudioChannel[CHANNEL0_PLAYBACK].Running==FALSE)
			{
				//
				//   Power down External amplifier and Headphone amplifier.
				//   Only if we are not playing digital audio
				//
				DURAUDIO_CodecWrite (pGeode, POWERDOWN_CTRL_STAT, DURAUDIO_PWR_D1);
			}

			pGeode->CurrentPowerState=DURAUDIO_D1;
			break;
		}

		case DURAUDIO_D2:
		case DURAUDIO_D3:
		case DURAUDIO_D4:
		{
			//
			// In this mode the power could be removed or reduced to a point
            // where the AC97 controller looses information, 
			// so we must save the codec registers.
            //
			DURAUDIO_SaveAudioContext(pGeode);

            //
            // Powerdown ADC, DAC, Mixer, HP amp, and Exernal Amp but not
            // AC-link and Clk
            //
			DURAUDIO_CodecWrite (pGeode, POWERDOWN_CTRL_STAT, DURAUDIO_PWR_D4);
			pGeode->CurrentPowerState=DURAUDIO_D4;

			break;
		}
	}

    return DURAUDIO_SUCCESS;
}

void DURAUDIO_DisableSMIs (PDURAUDIO pGeode)
{
	//
	//		Disable serial interrupts from the codec
	//
	if (pGeode->fIOAccess)
	{
		if (pGeode->fCS5535)
		{
		}
		else
		{
			OS_WritePortULong ( (unsigned short) (pGeode->F3BAR0 + 0x08), OS_ReadPortULong ((unsigned short) (pGeode->F3BAR0 + 0x08)) & 0xFF7FFFFF );
		}
	}
	else
	{
		*((unsigned long *)( pGeode->F3BAR0 + 0x08 )) = *((unsigned long *)( pGeode->F3BAR0 + 0x08 )) & 0xFF7FFFFF;
	}
}

void DURAUDIO_ResetCodec (PDURAUDIO pGeode)
{
	//
	//		Reset codec
	//
	if (pGeode->fIOAccess)
	{
		OS_WritePortULong ( (unsigned short) (pGeode->F3BAR0 + CODEC_CMD_REG), 0L );
	}
	else
	{
		*((unsigned long *)( pGeode->F3BAR0 + CODEC_CMD_REG )) = 0L;
	}
}

unsigned char DURAUDIO_CodecFullOn (PDURAUDIO pGeode)
{
	unsigned short	CodecData, Timeout;

	//
	//   Turn Everything ON
	//
	DURAUDIO_CodecWrite (pGeode, POWERDOWN_CTRL_STAT, DURAUDIO_PWR_D0);

	//
	//   Wait for the Codec to be powered up.
	//   Try it 10 times...
	//	
	Timeout=10;

	do
	{
		//
		//   Read the power management register.
		//
		CodecData=DURAUDIO_CodecRead (pGeode, POWERDOWN_CTRL_STAT);

		//
		//   Check the power state. Should be ready.
		//
		if ((CodecData & 0x000f) == 0x000f)
			break;
		else
			Timeout--;

		//
		// Let's wait a little, 10ms and then try again.
		//
		OS_Sleep(10);
	} while (Timeout);

	//
	//   Check if we timed out.
	//
	if (Timeout==0)
	{
		OS_DbgMsg ("PowerUpCodec timed out. CoDec not powered up.");
		return	FALSE;
	}

	return TRUE;
}


void DURAUDIO_SaveAudioContext (PDURAUDIO pGeode)
{
	unsigned char	i, RegIndex = 0;
	unsigned long	Channel;

	//
	//  Check if DMA is running for each channel.
	//  If it is, save PRD table address and turn it OFF
	//
	for (Channel=0;Channel<MAX_CHANNELS; Channel++)
	{
		if (pGeode->AudioChannel[Channel].Running==TRUE)
		{
			//
			//   Save Current PRD address
			//
			pGeode->AudioBusMaster_PRDTableAddress[Channel]  = *((unsigned long  *) (pGeode->F3BAR0+0x24+(Channel*0x08)));

			//
			//   Stop DMA
			//
			*((unsigned char *) pGeode->AudioChannel[Channel].AudioBusMaster.CommandRegister) = 
				*((unsigned char *) pGeode->AudioChannel[Channel].AudioBusMaster.CommandRegister) & (~ENABLE_BUSMASTER);
		}
	}
	//
	//		Save Mixer volumes and settings
	//
	for (i=0x02; i<=0x1E; i+=2)
	{
		pGeode->CODECRegisters[RegIndex++]=DURAUDIO_CodecRead(pGeode, i);
	}

	//
	//		Save Extended registers   (DAC/ADC rates etc...)
	//
	for (i=0x28; i<=0x38; i+=2)
	{
		pGeode->CODECRegisters[RegIndex++]=DURAUDIO_CodecRead(pGeode, i);
	}

	//
	//		Save F3BAR0 registers   
	//
	if (pGeode->fIOAccess)
	{
		//  TODO: Read all registers from the port and save them to pGeode->F3BARSave
	}
	else
	{
		memcpy ( pGeode->F3BARSave, (unsigned long *) pGeode->F3BAR0, 0x50 );
	}

	//
	//    Disable serial interrupts
	//
	if (pGeode->fIOAccess)
	{
		OS_WritePortULong ( (unsigned short) (pGeode->F3BAR0 + 0x08), OS_ReadPortULong ( (unsigned short) (pGeode->F3BAR0 + 0x08)) & 0xFF7FFFFF );
	}
	else
	{
		*((unsigned long *)( pGeode->F3BAR0 + 0x08 )) = *((unsigned long *)( pGeode->F3BAR0 + 0x08 )) & 0xFF7FFFFF;
	}
}

void DURAUDIO_RestoreAudioContext (PDURAUDIO pGeode)
{
	unsigned char	i, RegIndex = 0;
	unsigned long	Channel;

	//
	//		Restore F3BAR0 registers   
	//
	if (pGeode->fIOAccess)
	{
		//  TODO: send out to port F3BAR0 all the stored values for the registers
	}
	else
	{
		 memcpy ( (unsigned long *) pGeode->F3BAR0, pGeode->F3BARSave, 0x50 );

		//
		//  Check if DMA was running before suspend for each channel.
		//  If it was, restore the PRD table address and turn it back ON.
		//
		for (Channel=0;Channel<MAX_CHANNELS; Channel++)
		{
			if (pGeode->AudioChannel[Channel].Running==TRUE)
			{
				//
				//   Put back the PRD pointer
				//
				*(unsigned long  *) (pGeode->F3BAR0+0x24+(Channel*0x08)) = pGeode->AudioBusMaster_PRDTableAddress[Channel];

				//
				//   Start DMA
				//
				DURAUDIO_ResumeDMA  (pGeode, Channel);
			}
		}
	}

	//
	//		Restore Mixer volumes and settings
	//
	for (i=0x02; i<=0x1E; i+=2)
	{
		DURAUDIO_CodecWrite(pGeode, i, pGeode->CODECRegisters[RegIndex++]);
	}

	//
	//		Restore Extended registers   (DAC/ADC rates etc...)
	//
	for (i=0x28; i<=0x38; i+=2)
	{
		DURAUDIO_CodecWrite(pGeode, i, pGeode->CODECRegisters[RegIndex++]);
	}
}

//----------------------------------------------------------
// unsigned char DURAUDIO_ClearStat(PDURAUDIO pGeode, unsigned long Channel)
//
//  Clear the status bit of the respective Channel
//
//----------------------------------------------------------
unsigned char DURAUDIO_ClearStat(PDURAUDIO pGeode, unsigned long Channel)
{
	volatile unsigned char  status;     // Volatile to force read-to-clear.

	//	
	// Read to clear
	//
	if (pGeode->fIOAccess)
	{
		status = OS_ReadPortUChar ((unsigned short) pGeode->AudioChannel[Channel].AudioBusMaster.SMI_StatusRegister); 
	}
	else
	{
		status = *((unsigned char *) pGeode->AudioChannel[Channel].AudioBusMaster.SMI_StatusRegister);
	}

	return status;
}

//----------------------------------------------------------
//  void DURAUDIO_ClearIRQ(PDURAUDIO pGeode)
//
//  Clear the IRQ
//
//----------------------------------------------------------
void DURAUDIO_ClearIRQ (PDURAUDIO pGeode)
{
	volatile unsigned long  IRQStatus;   
	volatile unsigned short	wIRQStatus;

	if (pGeode->fCS5535)
	{
		DURAUDIO_ClearStat(pGeode, CHANNEL0_PLAYBACK);
		DURAUDIO_ClearStat(pGeode, CHANNEL1_RECORD);
	}
	else
	{
		//	
		// Read the IRQ Control register 
		//
		IRQStatus = *((unsigned long *) pGeode->IRQControlRegister);

		//
		//   Clear the respective IRQ bit 
		//
		IRQStatus = IRQStatus & (0xFFFF0000 | pGeode->uIRQMask);

		*((unsigned long *) pGeode->IRQControlRegister) = IRQStatus;

		//	
		// Read the Internal IRQ Enable register
		//
		wIRQStatus = *((unsigned short *) pGeode->InternalIRQEnableRegister);

		//
		//   Clear the respective IRQ bit - Set it back to external interrupt
		//
		wIRQStatus = wIRQStatus & pGeode->uIRQMask;

		*((unsigned short *) pGeode->InternalIRQEnableRegister) = wIRQStatus;
	}
}

//-------------------------------------------------------------------------
//   unsigned short DURAUDIO_IO_Read16(unsigned long pdev, unsigned short port)
//
//   Reads 16 bits to the PCI I/O space
//-------------------------------------------------------------------------
unsigned short DURAUDIO_IO_Read16(unsigned long pdev, unsigned short port)
{
	OS_WritePortULong (PCI_CADDR, pdev + port);
	
	return OS_ReadPortUShort (PCI_CDATA);
}

//------------------------------------------------------------------------------------
//   void DURAUDIO_IO_Write16(unsigned long pdev, unsigned short port, unsigned short data )
//
//   Writes 16 bits to the PCI I/O space
//------------------------------------------------------------------------------------
void DURAUDIO_IO_Write16(unsigned long pdev, unsigned short port, unsigned short data )
{
	OS_WritePortULong  (PCI_CADDR, pdev + port);
	
	OS_WritePortUShort (PCI_CDATA, data);
}

//-------------------------------------------------------------------------
//    unsigned long DURAUDIO_IO_Read32(unsigned long pdev, unsigned short port)
//
//   Reads 32 bits to the PCI I/O space
//-------------------------------------------------------------------------
unsigned long DURAUDIO_IO_Read32(unsigned long pdev, unsigned short port)
{
	OS_WritePortULong (PCI_CADDR, pdev + port);
	
	return (OS_ReadPortULong (PCI_CDATA));
}

//-----------------------------------------------------------------------------------
//    void DURAUDIO_IO_Write32(unsigned long pdev, unsigned short port, unsigned long data)
//
//   Writes 32 bits to the PCI I/O space
//-----------------------------------------------------------------------------------
void DURAUDIO_IO_Write32(unsigned long pdev, unsigned short port, unsigned long data)
{
	OS_WritePortULong (PCI_CADDR, pdev + port);
	
	OS_WritePortULong (PCI_CDATA, data );
}

//------------------------------------------------------------------------
//  unsigned long * DURAUDIO_GetDebuglevel (PDURAUDIO pGeode)
//
//  Returns the debug level that is stored on F3BAR+4C
//------------------------------------------------------------------------

unsigned long DURAUDIO_GetDebuglevel (PDURAUDIO pGeode)
{
	return *((unsigned long *)(pGeode->F3BAR0 + 0x4C));
}

//------------------------------------------------------------------------
//  unsigned long * DURAUDIO_GetDebuglevel (void)
//
//  Sets the debug level that is stored on F3BAR+4C
//------------------------------------------------------------------------
void DURAUDIO_SetDebuglevel (PDURAUDIO pGeode, unsigned long Level)
{
	*((unsigned long *)(pGeode->F3BAR0 + 0x4C))=Level;
}

//------------------------------------------------------------------------
//  unsigned long DURAUDIO_GetPositionInBuffer (PDURAUDIO pGeode, unsigned long Channel)
// 
//  Returns the position in bytes of the current data being played in the current Buffer. 
//------------------------------------------------------------------------
unsigned long DURAUDIO_GetPositionInBuffer (PDURAUDIO pGeode, unsigned long Channel, unsigned long PRDIndex)
{
	unsigned long	BeginningOfCurrentBuffer, DMACurrentPointer;
	unsigned char	CommandRegStatus;

	BeginningOfCurrentBuffer = pGeode->AudioChannel[Channel].PRDTable[PRDIndex].ulPhysAddr;

	if (pGeode->fIOAccess)
	{
		CommandRegStatus  = OS_ReadPortUChar ((unsigned short) pGeode->AudioChannel[Channel].AudioBusMaster.CommandRegister);
		DMACurrentPointer = OS_ReadPortULong ((unsigned short) pGeode->AudioChannel[Channel].AudioBusMaster.CurrentBufferPointer);
	}
	else
	{
		CommandRegStatus  = (unsigned char) pGeode->AudioChannel[Channel].AudioBusMaster.CommandRegister;
		DMACurrentPointer = (unsigned char) pGeode->AudioChannel[Channel].AudioBusMaster.CurrentBufferPointer;
	}

	//
	//    If the BusMaster is not enabled, return 0 as position
	//
	if (!( CommandRegStatus & ENABLE_BUSMASTER))
		return 0;

	
	return DMACurrentPointer-BeginningOfCurrentBuffer;
}

//------------------------------------------------------------------------
//  unsigned long DURAUDIO_GetPosition (PDURAUDIO pGeode, unsigned long Channel)
// 
//  Returns the position in bytes of the current data being played. 
//------------------------------------------------------------------------
unsigned long DURAUDIO_GetPosition (PDURAUDIO pGeode, unsigned long Channel)
{
	unsigned long	OffsetBytePointer, BytePosition, PRDIndex;

	PRDIndex=DURAUDIO_GetCurrentPRDIndex (pGeode, Channel);

	OffsetBytePointer = pGeode->AudioChannel[Channel].PRDTable[PRDIndex].ulPhysAddr;
	  
	//
	//   If we are right at the JMP PRD entry, Return 0 (beginning of the buffer)
	//
	if (OffsetBytePointer==pGeode->AudioChannel[Channel].PRDPhys)
		return 0;

	BytePosition=OffsetBytePointer - pGeode->AudioChannel[Channel].PRDTable[0].ulPhysAddr;
/*
	OS_DbgMsg("OffsetBytePointer:%08X\n", OffsetBytePointer);
	OS_DbgMsg("ulPhysAddr:%08X\n", pGeode->AudioChannel[Channel].PRDTable[0].ulPhysAddr);
	OS_DbgMsg("index on PRD:%d\n", PRDIndex);
	OS_DbgMsg("BytePosition:%d\n", BytePosition);
*/
	return BytePosition;    
}

//------------------------------------------------------------------------
//  unsigned long DURAUDIO_GetCurrentPRDIndex (PDURAUDIO pGeode, unsigned long Channel)
// 
//  Returns the Current PRD index being played
//------------------------------------------------------------------------
unsigned long DURAUDIO_GetCurrentPRDIndex (PDURAUDIO pGeode, unsigned long Channel)
{
	unsigned char	CommandRegStatus;
	unsigned long	PRDPointer, PRDIndex;

	//
	//    If the BusMaster is not enabled, return 0 as position
	//
	if (pGeode->fIOAccess)
	{
		CommandRegStatus = OS_ReadPortUChar ((unsigned short) pGeode->AudioChannel[Channel].AudioBusMaster.CommandRegister);
	}
	else
	{
		CommandRegStatus = *((unsigned char *)  pGeode->AudioChannel[Channel].AudioBusMaster.CommandRegister);
	}

	if (!( CommandRegStatus & ENABLE_BUSMASTER))
		return 0;

	if (pGeode->fIOAccess)
	{
		if (Channel==CHANNEL0_PLAYBACK)
			PRDPointer = OS_ReadPortULong ((unsigned short) (pGeode->F3BAR0 + 0x24) );    
		else
			PRDPointer = OS_ReadPortULong ((unsigned short) (pGeode->F3BAR0 + 0x2C) );
	}
	else
	{
		if (Channel==CHANNEL0_PLAYBACK)
			PRDPointer = *((unsigned long *) ( pGeode->F3BAR0 + 0x24 ));    
		else
			PRDPointer = *((unsigned long *) ( pGeode->F3BAR0 + 0x2C ));
	}

	if (PRDPointer==pGeode->AudioChannel[Channel].PRDPhys)
	{
		PRDIndex=0;
	}
	else
	{
		PRDIndex=(PRDPointer - pGeode->AudioChannel[Channel].PRDPhys) / sizeof(PRD_ENTRY) - 1;
	}


	return PRDIndex;
}

//------------------------------------------------------------------------
//  unsigned long DURAUDIO_DisplayCODECValues (PDURAUDIO pGeode)
//
//  Returns the position in bytes of the current data being played. 
//------------------------------------------------------------------------
void DURAUDIO_DisplayCODECValues (PDURAUDIO pGeode)
{
	unsigned char	i;
	unsigned short  Value;

	OS_DbgMsg("REG	Value");
	OS_DbgMsg("------------");

	for (i=0;i<0x7F;i+=2)
	{
		Value=DURAUDIO_CodecRead(pGeode, i);

		OS_DbgMsg("%04X", Value);
	}
}

//--------------------------------------------------------------------------------------------------
//  unsigned char DURAUDIO_AllocateDMA (PDURAUDIO pGeode, unsigned long Channel, unsigned long ulBufferSize)
//
//  Allocates the DMA buffer of ulBufferSize bytes for the specified Channel 
//--------------------------------------------------------------------------------------------------
unsigned char DURAUDIO_AllocateDMA (PDURAUDIO pGeode, unsigned long Channel, unsigned long ulBufferSize)
{
	pGeode->AudioChannel[Channel].DMA_VirtualAddress = (void *) OS_AllocateDMAMemory (	pGeode->pAdapterObject, 
																							ulBufferSize, 
																						   &pGeode->AudioChannel[Channel].DMA_PhysicalAddress);

    if (pGeode->AudioChannel[Channel].DMA_VirtualAddress == NULL) 
	{
		OS_DbgMsg("DURAUDIO_AllocateDMA - Unable to allocate %d bytes of memory for DMA buffer - Channel: %d\r\n", ulBufferSize, Channel);
        return FALSE;
    }

	return TRUE;
}

//--------------------------------------------------------------------------------------------------
//  unsigned char DURAUDIO_AllocatePRD (PDURAUDIO pGeode, unsigned long Channel, unsigned long Size)
//
//  Allocates the PRD table for a DMA buffer of DMABufferSize bytes for the specified Channel 
//--------------------------------------------------------------------------------------------------
unsigned char DURAUDIO_AllocatePRD (PDURAUDIO pGeode, 
									unsigned long Channel, 
									unsigned long DMABufferSize)
{
	unsigned long	PRDsToAllocate;

	OS_DbgMsg("Before OS_AllocateDMAMemory.\n");

	PRDsToAllocate=(DMABufferSize/SIZE_OF_EACH_PRD_BLOCK+1);
	pGeode->AudioChannel[Channel].PRDTable = (PPRD_ENTRY) OS_AllocateDMAMemory (pGeode->pAdapterObject, 
																			 PRDsToAllocate * sizeof(PRD_ENTRY), 
																			 &pGeode->AudioChannel[Channel].PRDPhys);

	OS_DbgMsg("After OS_AllocateDMAMemory.\n");
	if (!pGeode->AudioChannel[Channel].PRDTable)
		return FALSE;

	pGeode->AudioChannel[Channel].ulDMABufferSize		= DMABufferSize;
	pGeode->AudioChannel[Channel].PRDEntriesAllocated	= PRDsToAllocate;

	OS_DbgMsg("DURAUDIO_AllocatePRD: Returning.\n");

	return TRUE;
}

//--------------------------------------------------------------------------------------------------
//  void DURAUDIO_FreePRD (PDURAUDIO pGeode, unsigned long Channel)
//
//  Frees the PRD allocated by DURAUDIO_AllocatePRD
//--------------------------------------------------------------------------------------------------
void DURAUDIO_FreePRD (PDURAUDIO pGeode, unsigned long Channel)
{
	OS_FreeDMAMemory (  pGeode->pAdapterObject, 
						(void *) pGeode->AudioChannel[Channel].PRDTable, 
						pGeode->AudioChannel[Channel].PRDPhys, 
						pGeode->AudioChannel[Channel].PRDEntriesAllocated*8);
}


//--------------------------------------------------------------------------------------------------
//  void DURAUDIO_FreeDMA (PDURAUDIO pGeode, unsigned long Channel)
//
//  Frees the DMA allocated by DURAUDIO_AllocateDMA
//--------------------------------------------------------------------------------------------------
void DURAUDIO_FreeDMA (PDURAUDIO pGeode, unsigned long Channel)
{
	OS_FreeDMAMemory (  pGeode->pAdapterObject, 
						(void *) pGeode->AudioChannel[Channel].DMA_VirtualAddress, 
						pGeode->AudioChannel[Channel].DMA_PhysicalAddress, 
						pGeode->AudioChannel[Channel].ulDMABufferSize);
}


//---------------------------------------------------------------------------------------
//  void DURAUDIO_InitPRD  (PDURAUDIO	  pGeode, 
//							unsigned long Channel, 
//							unsigned long DMAPhys)
//
//  Initializes the PRD for the specified channel
//---------------------------------------------------------------------------------------
void DURAUDIO_InitPRD (PDURAUDIO	 pGeode, 
					   unsigned long Channel, 
					   unsigned long DMAPhys)
{
	//
	//    Fill the PRD Entries with the right sizes and flags=0
	//

	unsigned int	i;

	for (i=0; i<pGeode->AudioChannel[Channel].PRDEntriesAllocated; ++i)
	{
		pGeode->AudioChannel[Channel].PRDTable[i].ulPhysAddr = DMAPhys + (i * SIZE_OF_EACH_PRD_BLOCK);
		pGeode->AudioChannel[Channel].PRDTable[i].SizeFlags  = SIZE_OF_EACH_PRD_BLOCK;
	}

}

//------------------------------------------------------------------------
//	void DURAUDIO_SetupPRD(PDURAUDIO	  pGeode, 
//						   unsigned  long Channel, 
//						   unsigned  long SampleRate, 
//						   unsigned  long InterruptFrequency,
//						   unsigned  char FrequencyType)
//
//  Setup the PRD entries of the specified channel with EOP to generate 
//  interrupts based on the Sample rate and the InterruptFrequency 
//  value expressed in Miliseconds.
//
//------------------------------------------------------------------------
void DURAUDIO_SetupPRD(PDURAUDIO	 pGeode, 
					   unsigned long Channel, 
					   unsigned long SampleRate, 
					   unsigned long InterruptFrequency,
					   unsigned char FrequencyType)
{
	unsigned long	m_IntFreq, IntInc;
	unsigned int	i, Temp_i;

	if (FrequencyType==INTERVAL_BYTES)
	{
		m_IntFreq=(InterruptFrequency / SIZE_OF_EACH_PRD_BLOCK);
	}
	else
	{
		m_IntFreq=( ( (SampleRate * 4) * InterruptFrequency) / 1000 ) / SIZE_OF_EACH_PRD_BLOCK;
	}

	IntInc=m_IntFreq;

	if(pGeode->fPolling)
	{
		OS_DbgMsg("Build Version 3.0, POLLING\n" );
		OS_DbgMsg("PollTimeOut:%d  m_Freq:%d \n", InterruptFrequency, m_IntFreq );
	}
	else
	{
		OS_DbgMsg("Build Version 1.0, USING IRQ\n" );
		OS_DbgMsg("InterruptFrequency:%d  m_IntFreq:%d \n", InterruptFrequency, m_IntFreq );
	}

	OS_DbgMsg("PRDEntriesAllocated:%d\n", pGeode->AudioChannel[Channel].PRDEntriesAllocated );
	OS_DbgMsg("DMABufferSize:%d\n", pGeode->AudioChannel[Channel].ulDMABufferSize );

	for (i=0; i<=(pGeode->AudioChannel[Channel].PRDEntriesAllocated-1); ++i)
	{
		if (i == IntInc)
		{
			Temp_i=i;
			//
			//   If we are at the last entry, we put in the EOP in entry before
			//
			if (Temp_i == (pGeode->AudioChannel[Channel].PRDEntriesAllocated-1))
			{
				--Temp_i;
			}
			
			pGeode->AudioChannel[Channel].PRDTable[Temp_i].SizeFlags |= PRD_EOP_BIT;		// Set EOP bit
			IntInc+=m_IntFreq;
			OS_DbgMsg("PRDTable[%d] set to EOP\n", Temp_i );
		}
		else
			pGeode->AudioChannel[Channel].PRDTable[i].SizeFlags &= 0x0FFFFFFF;			// Clear Flags (keeps size)
	}

	//
	//	And, last, put the JMP to the top
	//
	pGeode->AudioChannel[Channel].PRDTable[pGeode->AudioChannel[Channel].PRDEntriesAllocated-1].ulPhysAddr = pGeode->AudioChannel[Channel].PRDPhys;
	pGeode->AudioChannel[Channel].PRDTable[pGeode->AudioChannel[Channel].PRDEntriesAllocated-1].SizeFlags  = PRD_JMP_BIT;  
}

//------------------------------------------------------------------------
//  void DURAUDIO_DisplayPRD(PDURAUDIO pGeode, unsigned long Channel)
//
//  Display the contents of the PRD table for the specific channel
//------------------------------------------------------------------------
void DURAUDIO_DisplayPRD(PDURAUDIO pGeode, unsigned long Channel)
{
	unsigned long		i;

	for (i=0; i<=(pGeode->AudioChannel[Channel].PRDEntriesAllocated); ++i)
	{
		OS_DbgMsg("%08X %08X\n",
					pGeode->AudioChannel[Channel].PRDTable[i].ulPhysAddr, 
					pGeode->AudioChannel[Channel].PRDTable[i].SizeFlags);
	}
}

//------------------------------------------------------------------------
//  unsigned char DURAUDIO_InterruptID (PDURAUDIO pGeode)
//
//  Returns the Interrupt Flags from the controller.
//	This is the mirror of the controller's status register
//
//	0x04	-	DMA0 interrupt (Wave Out)
//	0x08	-	DMA1 interrupt (Wave In)
//
//------------------------------------------------------------------------
unsigned char DURAUDIO_InterruptID (PDURAUDIO pGeode)
{
	volatile unsigned char    *TempInterruptID, ID;

	if (pGeode->fCS5535)
	{
		if (pGeode->fIOAccess)
		{
			ID=(unsigned char) OS_ReadPortUShort((unsigned short) (pGeode->F3BAR0 + 0x12));
		}
		else
		{
			TempInterruptID=(unsigned char *) (pGeode->F3BAR0 + 0x12);
			ID=*TempInterruptID;
		}
	}
	else
	{
		ID = *(pGeode->pInterruptID); 

		//
		//   Clear the Interrupt ID bits
		//
		*(pGeode->pInterruptID) = 0x00000000;

		//
		//   If there is no 0x4F0 mechanism, read it directly from the status register
		//
		if (ID==0)
		{
			if (pGeode->fIOAccess)
			{
				ID = (unsigned char) OS_ReadPortUShort ((unsigned short) (pGeode->F3BAR0 + 0x12));
			}
			else
			{
				TempInterruptID=(unsigned char *) (pGeode->F3BAR0 + 0x12);
				ID=*TempInterruptID;
			}
		}
	}

	return (ID);
}

//----------------------------------------------------
//
//  DURAUDIO_ReadStatus (PDURAUDIO pGeode, unsigned int BusMaster)
//
//  Reads the Status bits of the specified Bus Master
//
//----------------------------------------------------
unsigned char DURAUDIO_ReadStatus (PDURAUDIO pGeode, unsigned int BusMaster)
{
	if (pGeode->fIOAccess)
	{
		return OS_ReadPortUChar ((unsigned short) (pGeode->F3BAR0 + (0x21 + (8*BusMaster))));
	}
	else
	{
		return *((unsigned char *) (pGeode->F3BAR0 + 0x21 + (8*BusMaster)));
	}
}


//--------------------------------------------------------------
//
//				Indirect DMA Functions
//
//--------------------------------------------------------------

//--------------------------------------------------------------
//
//  FillBufferM8(PDURAUDIO pGeode,durPWAVEHDR pwh, unsigned char *pBuffer)
//
//	Fills the device's DMA buffer with Mono 8bit data 
//	from the user (Playback)
//
//--------------------------------------------------------------
void FillBufferM8(PDURAUDIO pGeode, durPWAVEHDR pwh, unsigned char *pBuffer)
{
    unsigned long	ulOffset;
    unsigned short	wData;

    ulOffset = 0;
    while (ulOffset < AUDIO_DMA_PAGE_SIZE) {
        if (pwh != NULL && pwh->dwBytesRecorded >= pwh->dwBufferLength) {
            pwh = pwh->lpNext;
            continue;
        }

        if (pwh != NULL) {
            wData = *(pwh->lpData + pwh->dwBytesRecorded);
            wData = (unsigned short)BITS_8_TO_16(wData);
            *((unsigned short *)(pBuffer + ulOffset)) = wData;
            *((unsigned short *)(pBuffer + ulOffset + 2)) = wData;
            ulOffset += 4;
            pwh->dwBytesRecorded += 1;
        } else {
            // nothing left to copy, clear remaining buffer
            memset(pBuffer + ulOffset, 0x80, AUDIO_DMA_PAGE_SIZE - ulOffset);
            ulOffset = AUDIO_DMA_PAGE_SIZE;
        }
    }
}

//--------------------------------------------------------------
//
//  FillBufferM16(PDURAUDIO pGeode,durPWAVEHDR pwh, unsigned char   *pBuffer)
//
//	Fills the device's DMA buffer with Mono 16bit data 
//	from the user (Playback)
//
//--------------------------------------------------------------
void FillBufferM16(PDURAUDIO pGeode, durPWAVEHDR pwh, unsigned char *pBuffer)
{
    unsigned long ulOffset;
    unsigned short wData;

    ulOffset = 0;
    while (ulOffset < AUDIO_DMA_PAGE_SIZE) {
        if (pwh != NULL && pwh->dwBytesRecorded >= pwh->dwBufferLength) {
            pwh = pwh->lpNext;
            continue;
        }

        if (pwh != NULL) {
            wData = *((unsigned short *)(pwh->lpData + pwh->dwBytesRecorded));
            *((unsigned short *)(pBuffer + ulOffset)) = wData;
            *((unsigned short *)(pBuffer + ulOffset + 2)) = wData;
            ulOffset += 4;
            pwh->dwBytesRecorded += 2;
        } else {
            // nothing left to copy, clear remaining buffer
            memset(pBuffer + ulOffset, 0x00, AUDIO_DMA_PAGE_SIZE - ulOffset);
            ulOffset = AUDIO_DMA_PAGE_SIZE;
        }
    }
}

//--------------------------------------------------------------
//
//  FillBufferS8(PDURAUDIO pGeode, durPWAVEHDR pwh, unsigned char   *pBuffer)
//
//	Fills the device's DMA buffer with Stereo 8bit data 
//	from the user (Playback)
//
//--------------------------------------------------------------
void FillBufferS8(PDURAUDIO pGeode,durPWAVEHDR pwh, unsigned char *pBuffer)
{
    unsigned long ulOffset;
    unsigned short wData;

    ulOffset = 0;
    while (ulOffset < AUDIO_DMA_PAGE_SIZE) {
        if (pwh != NULL && pwh->dwBytesRecorded >= pwh->dwBufferLength) {
            pwh = pwh->lpNext;
            continue;
        }

        if (pwh != NULL) {
            wData = *(pwh->lpData + pwh->dwBytesRecorded);
            wData = (unsigned short)BITS_8_TO_16(wData);
            *((unsigned short *)(pBuffer + ulOffset)) = wData;
            wData = *(pwh->lpData + pwh->dwBytesRecorded + 1);
            wData = (unsigned short)BITS_8_TO_16(wData);
            *((unsigned short *)(pBuffer + ulOffset + 2)) = wData;
            ulOffset += 4;
            pwh->dwBytesRecorded += 2;
        } else {
            // nothing left to copy, clear remaining buffer
            memset(pBuffer + ulOffset, 0x80, AUDIO_DMA_PAGE_SIZE - ulOffset);
            ulOffset = AUDIO_DMA_PAGE_SIZE;
        }
    }
}

//--------------------------------------------------------------
//
//  FillBufferS16(PDURAUDIO pGeode, durPWAVEHDR pwh, unsigned char   *pBuffer)
//
//	Fills the device's DMA buffer with Stereo 16bit data 
//	from the user (Playback)
//
//--------------------------------------------------------------
void FillBufferS16(PDURAUDIO pGeode, durPWAVEHDR pwh, unsigned char *pBuffer)
{
    unsigned long ulOffset;
    unsigned long ulBytesCopied;

    ulOffset = 0;
    ulBytesCopied = 0;
    while (ulOffset < AUDIO_DMA_PAGE_SIZE) {
        if (pwh != NULL && pwh->dwBytesRecorded >= pwh->dwBufferLength) {
            pwh = pwh->lpNext;
            continue;
        }

        if (pwh != NULL) {
            ulBytesCopied = pwh->dwBufferLength - pwh->dwBytesRecorded;
            if (ulBytesCopied > AUDIO_DMA_PAGE_SIZE - ulOffset)
                ulBytesCopied = AUDIO_DMA_PAGE_SIZE - ulOffset;

            memcpy(pBuffer + ulOffset, pwh->lpData + pwh->dwBytesRecorded, ulBytesCopied);
            pwh->dwBytesRecorded += ulBytesCopied;
        } else {
            // nothing left to copy, clear remaining buffer
            memset(pBuffer + ulOffset, 0x00, AUDIO_DMA_PAGE_SIZE - ulOffset);
            ulBytesCopied += AUDIO_DMA_PAGE_SIZE - ulOffset;
        }
        ulOffset += ulBytesCopied;
    }
}

//--------------------------------------------------------------
//
//  GetBufferM8(PDURAUDIO pGeode, durPWAVEHDR pwh, unsigned char   *pBuffer)
//
//	Gets a Mono 16bit buffer from the Audio device (Recording)
//
//--------------------------------------------------------------
void GetBufferM8(PDURAUDIO pGeode, durPWAVEHDR pwh, unsigned char   *pBuffer)
{
    unsigned long ulOffset;
    unsigned short wData;
    unsigned char Data;

    ulOffset = 0;
    while (ulOffset < AUDIO_DMA_PAGE_SIZE) {
        if (pwh != NULL && pwh->dwBytesRecorded >= pwh->dwBufferLength) {
            pwh = pwh->lpNext;
            continue;
        }

        if (pwh == NULL) {
            pGeode->AudioChannel[WAPI_IN].v_fMoreData = FALSE;
            break;
        }

        wData = *((unsigned short *)(pBuffer + ulOffset));
        Data = BITS_16_TO_8(wData);
        *(pwh->lpData + pwh->dwBytesRecorded) = Data;
        ulOffset += 4;
        pwh->dwBytesRecorded += 1;
    }
}

//--------------------------------------------------------------
//
//  GetBufferM16(PDURAUDIO pGeode, durPWAVEHDR pwh, unsigned char   *pBuffer)
//
//	Gets a Mono 16bit buffer from the Audio device (Recording)
//
//--------------------------------------------------------------
void GetBufferM16(PDURAUDIO pGeode, durPWAVEHDR pwh, unsigned char   *pBuffer)
{
    unsigned long ulOffset;
    unsigned short wData;

    ulOffset = 0;
    while (ulOffset < AUDIO_DMA_PAGE_SIZE) {
        if (pwh != NULL && pwh->dwBytesRecorded >= pwh->dwBufferLength) {
            pwh = pwh->lpNext;
            continue;
        }

        if (pwh == NULL) {
            pGeode->AudioChannel[WAPI_IN].v_fMoreData = FALSE;
            break;
        }

        wData = *((unsigned short *)(pBuffer + ulOffset));
        *((unsigned short *)(pwh->lpData + pwh->dwBytesRecorded)) = wData;
        ulOffset += 4;
        pwh->dwBytesRecorded += 2;
    }
}

//--------------------------------------------------------------
//
//  GetBufferS8(PDURAUDIO pGeode, durPWAVEHDR pwh, unsigned char   *pBuffer)
//
//	Gets a Stereo 8bit buffer from the Audio device (Recording)
//
//--------------------------------------------------------------
void GetBufferS8(PDURAUDIO pGeode, durPWAVEHDR pwh, unsigned char   *pBuffer)
{
    unsigned long ulOffset;
    unsigned short wData;
    unsigned char Data;

    ulOffset = 0;
    while (ulOffset < AUDIO_DMA_PAGE_SIZE) {
        if (pwh != NULL && pwh->dwBytesRecorded >= pwh->dwBufferLength) {
            pwh = pwh->lpNext;
            continue;
        }

        if (pwh == NULL) {
            pGeode->AudioChannel[WAPI_IN].v_fMoreData = FALSE;
            break;
        }

        wData = *((unsigned short *)(pBuffer + ulOffset));
        Data = BITS_16_TO_8(wData);
        *(pwh->lpData + pwh->dwBytesRecorded) = Data;
        wData = *((unsigned short *)(pBuffer + ulOffset + 2));
        Data = BITS_16_TO_8(wData);
        *(pwh->lpData + pwh->dwBytesRecorded + 1) = Data;
        ulOffset += 4;
        pwh->dwBytesRecorded += 2;
    }
}

//--------------------------------------------------------------
//
//  GetBufferS16(PDURAUDIO pGeode, durPWAVEHDR pwh, unsigned char   *pBuffer)
//
//	Gets a Stereo 16bit buffer from the Audio device (Recording)
//
//--------------------------------------------------------------
void GetBufferS16(PDURAUDIO pGeode, durPWAVEHDR pwh, unsigned char *pBuffer)
{
    unsigned long ulOffset;
    unsigned long ulBytesCopied;

    ulOffset = 0;
    ulBytesCopied = 0;
    while (ulOffset < AUDIO_DMA_PAGE_SIZE) {
        if (pwh != NULL && pwh->dwBytesRecorded >= pwh->dwBufferLength) {
            pwh = pwh->lpNext;
            continue;
        }

        if (pwh == NULL) {
            pGeode->AudioChannel[WAPI_IN].v_fMoreData = FALSE;
            break;
        }

        ulBytesCopied = pwh->dwBufferLength - pwh->dwBytesRecorded;
        if (ulBytesCopied > AUDIO_DMA_PAGE_SIZE - ulOffset)
            ulBytesCopied = AUDIO_DMA_PAGE_SIZE - ulOffset;

        memcpy(pwh->lpData + pwh->dwBytesRecorded, pBuffer + ulOffset, ulBytesCopied);
        pwh->dwBytesRecorded += ulBytesCopied;
        ulOffset += ulBytesCopied;
    }
}

// -----------------------------------------------------------------------------
//   void DURAUDIO_WaveClose(PDURAUDIO pGeode, WAPI_INOUT Channel)
//
//   Closes the Wave Stream.
//   Sets the "In use" flag to FALSE;
// 
// -----------------------------------------------------------------------------
void DURAUDIO_WaveClose(PDURAUDIO pGeode, WAPI_INOUT Channel)
{
    OS_DbgMsg("+DURAUDIO_WaveClose");

    // Reset the digital volume to zero
    pGeode->v_nVolume = 0;

	pGeode->AudioChannel[Channel].fInUse=FALSE;

    OS_DbgMsg("-DURAUDIO_WaveClose");
};

void DURAUDIO_SetupDMA (PDURAUDIO pGeode, unsigned long Channel, unsigned long dma_page_size)
{   
   //
   //	Setup the DMA pages
   //
    pGeode->AudioChannel[Channel].IndirectDMA.dma_page[1] = pGeode->AudioChannel[Channel].IndirectDMA.dma_page[0] + dma_page_size; 
    pGeode->AudioChannel[Channel].IndirectDMA.dma_page[2] = pGeode->AudioChannel[Channel].IndirectDMA.dma_page[1] + dma_page_size; 
    pGeode->AudioChannel[Channel].IndirectDMA.dma_page[3] = pGeode->AudioChannel[Channel].IndirectDMA.dma_page[2] + dma_page_size; 

    //
    //  The 5530/5540 contain an "audio buffered PCI bus master" that performs DMA-like
    //  functions.
    //  The Physical Region Descriptor (PRD) specifies a buffer to be "DMAed" by the
    //  PCI bus master. The prd_array_phys_adr in the mem-mapped 5530 audio regs points
    //  to the start of an array of PRD descriptors. 
    //
    //  Each entry of this array consists of 2 32-bit words:
    //
    //   32-bit word 0:   Address of the DMA buffer  .
    //               1:   bit-31:   End of Table (EOT)
    //                    bit-30:   End of Page  (EOP)
    //                    bit-29:   Loop         (JMP)
    //                    bit-15:0  byte count.
    //
    //
    //  Here, the PRD table is initialized as following:
    //
    //                +----------------------------------------+
    //  prd_array[0]  !       Address of the 1st buffer        +
    //                +----------------------------------------+
    //  prd_array[1]  !  EOP bit set  !  Length of 1st Buffer  +
    //                +----------------------------------------+
    //  prd_array[2]  !       Address of the 2nd Buffer        +
    //                +----------------------------------------+
    //  prd_array[3]  !  EOP bit set  !  Length of 2nd Buffer  +
    //                +----------------------------------------+
    //  prd_array[4]  !       Pointer to 1st PRD entry         +
    //                +----------------------------------------+
    //  prd_array[5]  !  JMP bit set  !                        +
    //                +----------------------------------------+
    //
	
    //
    //  Logical address of the PRD 
    //
    pGeode->AudioChannel[Channel].IndirectDMA.prd_array = (unsigned long *)( pGeode->AudioChannel[Channel].IndirectDMA.dma_page[2] );     
	
    //
    //  Physical address of the PRD 
    //
    pGeode->AudioChannel[Channel].IndirectDMA.prd_array_phys_adr = pGeode->AudioChannel[Channel].DMA_PhysicalAddress + (2 * dma_page_size);  
	
    //
    // Address of the 1st buffer 
    //
    pGeode->AudioChannel[Channel].IndirectDMA.prd_array[0]  = pGeode->AudioChannel[Channel].DMA_PhysicalAddress;
	
    //
    // EOP flag  | Length of 1st Buffer
    //
    pGeode->AudioChannel[Channel].IndirectDMA.prd_array[1]  = (0x40000000 | (dma_page_size) );    
	
    //
    // Address of the 2nd Buffer
    //
    pGeode->AudioChannel[Channel].IndirectDMA.prd_array[2]  = pGeode->AudioChannel[Channel].DMA_PhysicalAddress + dma_page_size;
	
    //
    // EOP flag  | Length of 1st Buffer
    //
    pGeode->AudioChannel[Channel].IndirectDMA.prd_array[3]  = (0x40000000 | (dma_page_size) );      
	
    //
    // Pointer to the 1st PRD entry
    //
    pGeode->AudioChannel[Channel].IndirectDMA.prd_array[4]  = pGeode->AudioChannel[Channel].IndirectDMA.prd_array_phys_adr;
	
    //
    //  JMP flag - Loop to the beggining of the PRD
    //
    pGeode->AudioChannel[Channel].IndirectDMA.prd_array[5]  = 0x20000000;   
}

void DURAUDIO_Standby(PDURAUDIO pGeode, unsigned long Channel)
{
	DURAUDIO_ClearStat(pGeode, Channel);
	pGeode->AudioChannel[Channel].Running=FALSE;    
	pGeode->AudioChannel[Channel].fInUse =FALSE;   
}

//---------------------------------------------------------
//  void DURAUDIO_Free(PDURAUDIO pGeode, unsigned long Channel)
//
//  This routine is called when the driver 
//  is being unloaded. Free up memory, etc.
//---------------------------------------------------------
void DURAUDIO_Free(PDURAUDIO pGeode, unsigned long Channel)
{
	OS_Free(pGeode->AudioChannel[Channel].IndirectDMA.dma_page[0]);
}

//---------------------------------------------------------
//   unsigned long DURAUDIO_GetVolume(PDURAUDIO pGeode)
//  
//   Returns the digital volume to the caller
//---------------------------------------------------------
unsigned long DURAUDIO_GetVolume(PDURAUDIO pGeode)
{
	return pGeode->v_nVolume;
}

//---------------------------------------------------------
//   void DURAUDIO_SetVolume(PDURAUDIO pGeode, unsigned long Vol)
//
//   Sets the digital volume based on the value sent by the caller
//---------------------------------------------------------
void DURAUDIO_SetVolume(PDURAUDIO pGeode, unsigned long Vol)
{
    unsigned short rVol, lVol, nVol;
    unsigned char rIdx, lIdx;

    pGeode->v_nVolume = Vol;

    rIdx = (unsigned char)((Vol & 0xF8000000) >> (32 - 5));
    lIdx = (unsigned char)((Vol & 0x0000F800) >> (16 - 5));
    rVol = VolumeLUT[rIdx];
    lVol = VolumeLUT[lIdx];
    nVol = rVol | (lVol << 8);
    if (nVol == 0x1f1f)
        nVol |= 0x8000;

    DURAUDIO_CodecWrite(pGeode, MASTER_VOLUME, nVol);
}

// -----------------------------------------------------------------------------
//
//   AUDIO_STATE DURAUDIO_GetInterruptType(VOID)
//
//   Called by the driver's ISR.
//   Returns the state of the stream being Played/Recorded.
// -----------------------------------------------------------------------------
AUDIO_STATE DURAUDIO_GetInterruptType(PDURAUDIO pGeode)
{
    AUDIO_STATE asResult = AUDIO_STATE_IGNORE;
    volatile unsigned long * pPRDAddr;
    volatile unsigned long ulNextPRD;

    // check WAPI_IN first
    if (pGeode->AudioChannel[WAPI_IN].Running) 
	{
        if (pGeode->AudioChannel[WAPI_IN].v_fMoreData) 
		{
            pPRDAddr = (unsigned long *)(pGeode->F3BAR0 + 0x2C);
            ulNextPRD = *pPRDAddr;

            if (ulNextPRD != pGeode->AudioChannel[WAPI_IN].LastAddr) 
			{
                pGeode->AudioChannel[WAPI_IN].LastAddr = ulNextPRD;
                asResult |= AUDIO_STATE_IN_RECORDING;
            }
        } 
		else 
		{
            asResult |= AUDIO_STATE_IN_STOPPED;
        }
    }

    // check WAPI_OUT
    if (pGeode->AudioChannel[WAPI_OUT].Running) 
	{
        if (pGeode->AudioChannel[WAPI_OUT].v_fMoreData) 
		{
            pPRDAddr = (unsigned long *)(pGeode->F3BAR0 + 0x24);
            ulNextPRD = *pPRDAddr;
            if (ulNextPRD != pGeode->AudioChannel[WAPI_OUT].LastAddr) 
			{
                pGeode->AudioChannel[WAPI_OUT].LastAddr = ulNextPRD;
                asResult |= AUDIO_STATE_OUT_PLAYING;
            }
        } 
		else 
		{
            asResult |= AUDIO_STATE_OUT_STOPPED;
        }
    }

    return asResult;
}




